"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DrawAndMatchItem = exports.DrawAndMatch = exports.DrawAndMatchDirection = void 0;
var DrawAndMatchDirection;
(function (DrawAndMatchDirection) {
    DrawAndMatchDirection[DrawAndMatchDirection["NONE"] = 0] = "NONE";
    DrawAndMatchDirection[DrawAndMatchDirection["RIGHT"] = 1] = "RIGHT";
    DrawAndMatchDirection[DrawAndMatchDirection["DOWN"] = 2] = "DOWN";
    DrawAndMatchDirection[DrawAndMatchDirection["LEFT"] = 4] = "LEFT";
    DrawAndMatchDirection[DrawAndMatchDirection["UP"] = 8] = "UP";
})(DrawAndMatchDirection = exports.DrawAndMatchDirection || (exports.DrawAndMatchDirection = {}));
var DrawAndMatch = /** @class */ (function () {
    function DrawAndMatch(options) {
        this.config = {
            rows: (options === undefined || options.rows === undefined) ? DrawAndMatch.DEFALUT_VALUES.rows : options.rows,
            columns: (options === undefined || options.columns === undefined) ? DrawAndMatch.DEFALUT_VALUES.columns : options.columns,
            items: (options === undefined || options.items === undefined) ? DrawAndMatch.DEFALUT_VALUES.items : options.items,
            tileSize: (options === undefined || options.tileSize === undefined) ? DrawAndMatch.DEFALUT_VALUES.tileSize : options.tileSize,
            startX: (options === undefined || options.startX === undefined) ? DrawAndMatch.DEFALUT_VALUES.startX : options.startX,
            startY: (options === undefined || options.startY === undefined) ? DrawAndMatch.DEFALUT_VALUES.startY : options.startY,
            insideEnough: (options === undefined || options.insideEnough === undefined) ? DrawAndMatch.DEFALUT_VALUES.insideEnough : options.insideEnough,
            minChain: (options === undefined || options.minChain === undefined) ? DrawAndMatch.DEFALUT_VALUES.minChain : options.minChain
        };
        this.gameArray = [];
        for (var i = 0; i < this.config.rows; i++) {
            this.gameArray[i] = [];
            for (var j = 0; j < this.config.columns; j++) {
                var randomValue = Math.floor(Math.random() * this.config.items);
                this.gameArray[i][j] = {
                    empty: false,
                    value: randomValue,
                    item: new DrawAndMatchItem(i, j, randomValue, this.config.startX + j * this.config.tileSize, this.config.startY + i * this.config.tileSize, this.config.tileSize)
                };
            }
        }
        this.chain = [];
        this.previousChainLenght = 0;
    }
    Object.defineProperty(DrawAndMatch.prototype, "items", {
        /* get all game items */
        get: function () {
            var items = [];
            for (var i = 0; i < this.config.rows; i++) {
                for (var j = 0; j < this.config.columns; j++) {
                    items.push(this.gameArray[i][j].item);
                }
            }
            return items;
        },
        enumerable: false,
        configurable: true
    });
    /* output the chain in the console */
    DrawAndMatch.prototype.logChain = function () {
        var _this = this;
        var output = '';
        this.chain.forEach(function (coordinate, index) {
            output += '[' + coordinate.row + ', ' + coordinate.column + ']';
            if (index < _this.chain.length - 1) {
                output += ' -> ';
            }
        });
        console.log(output);
    };
    /* output board values in console */
    DrawAndMatch.prototype.logBoard = function () {
        var output = '';
        for (var i = 0; i < this.config.rows; i++) {
            for (var j = 0; j < this.config.columns; j++) {
                output += (this.gameArray[i][j].empty ? '.' : this.gameArray[i][j].value) + ' ';
            }
            output += '\n';
        }
        console.log(output);
    };
    /* check a pick is in a valid row and column */
    DrawAndMatch.prototype.validPick = function (row, column) {
        return row >= 0 && row < this.config.rows && column >= 0 && column < this.config.columns && this.gameArray[row] != undefined && this.gameArray[row][column] != undefined;
    };
    /* check if input is inside game board */
    DrawAndMatch.prototype.isInputInsideBoard = function (x, y) {
        var column = Math.floor((x - this.config.startX) / this.config.tileSize);
        var row = Math.floor((y - this.config.startY) / this.config.tileSize);
        return this.validPick(row, column);
    };
    /* handle input movement */
    DrawAndMatch.prototype.handleInputMovement = function (x, y) {
        var items = [];
        if (this.isInputInsideBoard(x, y)) {
            if (this.isInsideEnough(x, y)) {
                if (this.isCoordinateInChain(x, y)) {
                    if (this.isBacktrack(x, y)) {
                        var removedTile = this.chain.pop();
                        var removedItem = this.gameArray[removedTile.row][removedTile.column].item;
                        removedItem.selected = false;
                        removedItem.arrowVisible = false;
                        items.push(removedItem);
                        var lastTile = this.chain[this.chain.length - 1];
                        var lastItem = this.gameArray[lastTile.row][lastTile.column].item;
                        lastItem.selected = true;
                        lastItem.arrowVisible = false;
                        items.push(lastItem);
                    }
                }
                else {
                    if (this.canContinueChain(x, y) && this.isChainValue(x, y)) {
                        this.chain.push(this.getCoordinateAt(x, y));
                        var addedTile = this.chain[this.chain.length - 1];
                        var addedItem = this.gameArray[addedTile.row][addedTile.column].item;
                        addedItem.selected = true;
                        addedItem.arrowVisible = false;
                        items.push(addedItem);
                        var formerLastTile = this.chain[this.chain.length - 2];
                        var formerLastItem = this.gameArray[formerLastTile.row][formerLastTile.column].item;
                        formerLastItem.selected = false;
                        formerLastItem.arrowVisible = true;
                        formerLastItem.arrowDirection = this.getArrowDirection(this.chain.length - 2);
                        items.push(formerLastItem);
                    }
                }
            }
        }
        return items;
    };
    /* get arrow direction */
    DrawAndMatch.prototype.getArrowDirection = function (index) {
        var coordinate1 = this.chain[index + 1];
        var coordinate2 = this.chain[index];
        var deltaRow = coordinate1.row - coordinate2.row;
        var deltaColumn = coordinate1.column - coordinate2.column;
        var direction = 0;
        direction += (deltaColumn < 0) ? DrawAndMatchDirection.LEFT : ((deltaColumn > 0) ? DrawAndMatchDirection.RIGHT : 0);
        direction += (deltaRow < 0) ? DrawAndMatchDirection.UP : ((deltaRow > 0) ? DrawAndMatchDirection.DOWN : 0);
        return direction;
    };
    /* check if a movement coordinate is a backtrack */
    DrawAndMatch.prototype.isBacktrack = function (x, y) {
        if (this.chain.length > 1) {
            var currentCoordinate = this.getCoordinateAt(x, y);
            var backtrackCoordinate = this.getChainBacktrack();
            return currentCoordinate.row == backtrackCoordinate.row && currentCoordinate.column == backtrackCoordinate.column;
        }
        return false;
    };
    /* get the chain head */
    DrawAndMatch.prototype.getChainHead = function () {
        return this.chain[this.chain.length - 1];
    };
    /* get chain backtrack item */
    DrawAndMatch.prototype.getChainBacktrack = function () {
        return this.chain[this.chain.length - 2];
    };
    /* check if a coordinate can continue the chain */
    DrawAndMatch.prototype.canContinueChain = function (x, y) {
        var currentCoordinate = this.getCoordinateAt(x, y);
        var previusCoordinate = this.getChainHead();
        var rowDifference = Math.abs(currentCoordinate.row - previusCoordinate.row);
        var columnDifference = Math.abs(currentCoordinate.column - previusCoordinate.column);
        return (rowDifference + columnDifference == 1) || (rowDifference == 1 && columnDifference == 1);
    };
    /* check if a coordinate matches chain value */
    DrawAndMatch.prototype.isChainValue = function (x, y) {
        var coordinate = this.getCoordinateAt(x, y);
        return this.gameArray[coordinate.row][coordinate.column].value == this.chainValue;
    };
    /* start the chain */
    DrawAndMatch.prototype.startChain = function (x, y) {
        if (this.isInputInsideBoard(x, y)) {
            var coordinate = this.getCoordinateAt(x, y);
            this.chain.push(coordinate);
            this.chainValue = this.gameArray[coordinate.row][coordinate.column].value;
            return this.gameArray[coordinate.row][coordinate.column].item;
        }
        return null;
    };
    /* check if a coordinate is in the chain */
    DrawAndMatch.prototype.isCoordinateInChain = function (x, y) {
        var result = false;
        var currentCoordinate = this.getCoordinateAt(x, y);
        this.chain.forEach(function (coordinate) {
            if (currentCoordinate.row == coordinate.row && currentCoordinate.column == coordinate.column) {
                result = true;
            }
        });
        return result;
    };
    /* check if a coordinate is inside enough to be considered an actual item selection */
    DrawAndMatch.prototype.isInsideEnough = function (x, y) {
        var coordinate = this.getCoordinateAt(x, y);
        var centerX = coordinate.column * this.config.tileSize + this.config.startX + this.config.tileSize / 2;
        var centerY = coordinate.row * this.config.tileSize + this.config.startY + this.config.tileSize / 2;
        var distanceX = centerX - x;
        var distanceY = centerY - y;
        var maxDistance = this.config.insideEnough * this.config.tileSize;
        return distanceX * distanceX + distanceY * distanceY <= maxDistance * maxDistance;
    };
    /* get the screen coordinate, given a draw and match coordinate */
    DrawAndMatch.prototype.getCoordinateAt = function (x, y) {
        return {
            row: Math.floor((y - this.config.startY) / this.config.tileSize),
            column: Math.floor((x - this.config.startX) / this.config.tileSize)
        };
    };
    /* remove items */
    DrawAndMatch.prototype.removeItems = function () {
        var _this = this;
        this.itemPool = [];
        var items = [];
        this.chain.forEach(function (coordinate) {
            _this.gameArray[coordinate.row][coordinate.column].item.toBeRemoved = _this.chain.length >= _this.config.minChain;
            items.push(_this.gameArray[coordinate.row][coordinate.column].item);
            if (_this.chain.length >= _this.config.minChain) {
                _this.gameArray[coordinate.row][coordinate.column].empty = true;
                _this.itemPool.push(_this.gameArray[coordinate.row][coordinate.column].item);
            }
        });
        this.chain = [];
        return items;
    };
    /* calculate how many empty spaces there are below (row, column) */
    DrawAndMatch.prototype.emptySpacesBelow = function (row, column) {
        var result = 0;
        if (row != this.config.rows) {
            for (var i = row + 1; i < this.config.rows; i++) {
                if (this.gameArray[i][column].empty) {
                    result++;
                }
            }
        }
        return result;
    };
    /* arrange the board after a draw */
    DrawAndMatch.prototype.arrangeBoard = function () {
        var result = [];
        for (var i = this.config.rows - 2; i >= 0; i--) {
            for (var j = 0; j < this.config.columns; j++) {
                var emptySpaces = this.emptySpacesBelow(i, j);
                if (!this.gameArray[i][j].empty && emptySpaces > 0) {
                    var item = this.gameArray[i][j].item;
                    item.row += emptySpaces;
                    var movement = new DrawAndMatchMovement(item.posX, item.posY, item.posX, item.posY + emptySpaces * this.config.tileSize, emptySpaces, this.config.tileSize);
                    item.posY += emptySpaces * this.config.tileSize;
                    item.movement = movement;
                    result.push(item);
                    var tempTile = this.gameArray[i + emptySpaces][j];
                    this.gameArray[i + emptySpaces][j] = this.gameArray[i][j];
                    this.gameArray[i][j] = tempTile;
                }
            }
        }
        for (var i = 0; i < this.config.columns; i++) {
            if (this.gameArray[0][i].empty) {
                var emptySpaces = this.emptySpacesBelow(0, i) + 1;
                for (var j = emptySpaces - 1; j >= 0; j--) {
                    var randomValue = Math.floor(Math.random() * this.config.items);
                    var item = this.itemPool.shift();
                    item.value = randomValue;
                    item.row = j;
                    item.column = i;
                    var movement = new DrawAndMatchMovement(this.config.startX + i * this.config.tileSize, this.config.startY - (emptySpaces - j) * this.config.tileSize, this.config.startX + i * this.config.tileSize, this.config.startY + j * this.config.tileSize, emptySpaces, this.config.tileSize);
                    item.posY = this.config.startY + j * this.config.tileSize;
                    item.posX = this.config.startX + i * this.config.tileSize;
                    item.movement = movement;
                    result.push(item);
                    this.gameArray[j][i].item = item;
                    this.gameArray[j][i].value = randomValue;
                    this.gameArray[j][i].empty = false;
                }
            }
        }
        return result;
    };
    DrawAndMatch.DEFALUT_VALUES = {
        rows: 6,
        columns: 8,
        items: 6,
        tileSize: 100,
        startX: 0,
        startY: 0,
        insideEnough: 0.4,
        minChain: 3
    };
    return DrawAndMatch;
}());
exports.DrawAndMatch = DrawAndMatch;
var DrawAndMatchItem = /** @class */ (function () {
    function DrawAndMatchItem(row, column, value, posX, posY, size) {
        this.row = row;
        this.column = column;
        this.value = value;
        this.posX = posX;
        this.posY = posY;
        this.size = size;
        this.centerX = posX + size / 2;
        this.centerY = posY + size / 2;
        this.selected = false;
        this.arrowVisible = false;
        this.arrowDirection = DrawAndMatchDirection.NONE;
    }
    DrawAndMatchItem.prototype.setData = function (data) {
        this.data = data;
    };
    return DrawAndMatchItem;
}());
exports.DrawAndMatchItem = DrawAndMatchItem;
var DrawAndMatchMovement = /** @class */ (function () {
    function DrawAndMatchMovement(startX, startY, endX, endY, deltaRow, size) {
        this.startX = startX;
        this.startY = startY;
        this.startCenterX = startX + size / 2;
        this.startCenterY = startY + size / 2;
        this.endX = endX;
        this.endY = endY;
        this.endCenterX = endX + size / 2;
        this.endCenterY = endY + size / 2;
        this.deltaRow = deltaRow;
    }
    return DrawAndMatchMovement;
}());
